﻿using HeuristicLab.Algorithms.ALPS;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Selection;

namespace HeuristicLab.Algorithms.IslandALPS {
  [Item("IslandALPSAlgorithmMainLoop", "An island ALPS main loop operator.")]
  [StorableClass]
  public class IslandAlpsAlgorthmsMainLoop : AlgorithmOperator{

    [StorableConstructor]
    private IslandAlpsAlgorthmsMainLoop(bool deserializing) : base(deserializing) { }

    #region Parameter Properties
    public ValueLookupParameter<IRandom> GlobalRandomParameter {
      get { return (ValueLookupParameter<IRandom>)Parameters["GlobalRandom"]; }
    }
    public ValueLookupParameter<IRandom> IslandRandomParameter {
      get { return (ValueLookupParameter<IRandom>)Parameters["IslandRandom"]; }
    }
    public ValueLookupParameter<IRandom> LayerRandomParameter {
      get { return (ValueLookupParameter<IRandom>)Parameters["LayerRandom"]; }
    }

    public ValueLookupParameter<BoolValue> MaximizationParameter {
      get { return (ValueLookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ValueLookupParameter<DoubleValue> BestKnownQualityParameter {
      get { return (ValueLookupParameter<DoubleValue>)Parameters["BestKnownQuality"]; }
    }
    
    public ValueLookupParameter<IntValue> PopulationSizeParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["PopulationSize"]; }
    }

    public ValueLookupParameter<IOperator> SelectorParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["Selector"]; }
    }
    public ValueLookupParameter<IOperator> CrossoverParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["Crossover"]; }
    }
    public ValueLookupParameter<PercentValue> MutationProbabilityParameter {
      get { return (ValueLookupParameter<PercentValue>)Parameters["MutationProbability"]; }
    }
    public ValueLookupParameter<IOperator> MutatorParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["Mutator"]; }
    }
    public ValueLookupParameter<IOperator> EvaluatorParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["Evaluator"]; }
    }
    public ValueLookupParameter<IntValue> ElitesParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["Elites"]; }
    }
    public IValueLookupParameter<BoolValue> ReevaluateElitesParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["ReevaluateElites"]; }
    }
    public ValueLookupParameter<ResultCollection> ResultsParameter {
      get { return (ValueLookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public ILookupParameter<ITerminator> TerminatorParameter {
      get { return (ILookupParameter<ITerminator>)Parameters["Terminator"]; }
    }

    public ValueLookupParameter<IOperator> AnalyzerParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["Analyzer"]; }
    }
    public ValueLookupParameter<IOperator> IslandAnalyzerParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["IslandAnalyzer"]; }
    }
    public ValueLookupParameter<IOperator> LayerAnalyzerParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["LayerAnalyzer"]; }
    }

    public ValueLookupParameter<IOperator> GroupAnalyzerParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["GroupAnalyzer"]; }
    }

    public LookupParameter<IntValue> LayerEvaluatedSolutionsParameter {
      get { return (LookupParameter<IntValue>)Parameters["LayerEvaluatedSolutions"]; }
    }
    public LookupParameter<IntValue> EvaluatedSolutions {
      get { return (LookupParameter<IntValue>)Parameters["EvaluatedSolutions"]; }
    }
    #region island
    public ValueLookupParameter<IntValue> NumberOfIslandsParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["NumberOfIslands"]; }
    }
    public ValueLookupParameter<IntValue> MigrationIntervalParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["MigrationInterval"]; }
    }
    public ValueLookupParameter<PercentValue> MigrationRateParameter {
      get { return (ValueLookupParameter<PercentValue>)Parameters["MigrationRate"]; }
    }
    public ValueLookupParameter<IOperator> MigratorParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["Migrator"]; }
    }
    public ValueLookupParameter<IOperator> EmigrantsSelectorParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["EmigrantsSelector"]; }
    }
    public ValueLookupParameter<IOperator> ImmigrationReplacerParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["ImmigrationReplacer"]; }
    }
    public ValueLookupParameter<BoolValue> Migrate {
      get { return (ValueLookupParameter<BoolValue>)Parameters["Migrate"]; }
    }
    public LookupParameter<IntValue> IslandEvaluatedSolutions {
      get { return (LookupParameter<IntValue>)Parameters["IslandEvaluatedSolutions"]; }
    }

    #endregion

    #region alps
    public ILookupParameter<IntValue> CurrentPopulationSizeParameter {
      get { return (ILookupParameter<IntValue>)Parameters["CurrentPopulationSize"]; }
    }
    public IValueLookupParameter<IntValue> NumberOfLayersParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["NumberOfLayers"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> AgeParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["Age"]; }
    }
    public IValueLookupParameter<IntValue> AgeGapParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["AgeGap"]; }
    }
    public IValueLookupParameter<DoubleValue> AgeInheritanceParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["AgeInheritance"]; }
    }
    public IValueLookupParameter<IntArray> AgeLimitsParameter {
      get { return (IValueLookupParameter<IntArray>)Parameters["AgeLimits"]; }
    }
    public IValueLookupParameter<IntValue> MatingPoolRangeParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MatingPoolRange"]; }
    }
    public IValueLookupParameter<BoolValue> ReduceToPopulationSizeParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["ReduceToPopulationSize"]; }
    }
    public IValueLookupParameter<BoolValue> PlusSelectionParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["PlusSelection"]; }
    }
    #endregion

    #endregion

    private IslandAlpsAlgorthmsMainLoop(IslandAlpsAlgorthmsMainLoop original, Cloner cloner) :
      base(original,cloner){ }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new IslandAlpsAlgorthmsMainLoop(this,cloner);
    }

    public IslandAlpsAlgorthmsMainLoop() : base() {

      #region Create parameters
      Parameters.Add(new ValueLookupParameter<IRandom>("GlobalRandom", "A pseudo random number generator."));
      Parameters.Add(new ValueLookupParameter<IRandom>("IslandRandom", "A pseudo random number generator."));
      Parameters.Add(new ValueLookupParameter<IRandom>("LayerRandom", "A pseudo random number generator."));

      Parameters.Add(new ValueLookupParameter<BoolValue>("Maximization", "True if the problem is a maximization problem, otherwise false."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", "The value which represents the quality of a solution."));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("BestKnownQuality", "The best known quality value found so far."));

      Parameters.Add(new ValueLookupParameter<IntValue>("NumberOfLayers", "The number of Layers per Island."));
      Parameters.Add(new ValueLookupParameter<IntValue>("NumberOfIslands", "The number of islands."));
      Parameters.Add(new ValueLookupParameter<IntValue>("MigrationInterval", "The number of generations that should pass between migration phases."));
      Parameters.Add(new ValueLookupParameter<PercentValue>("MigrationRate", "The proportion of individuals that should migrate between the islands."));
      Parameters.Add(new ValueLookupParameter<IOperator>("Migrator", "The migration strategy."));
      Parameters.Add(new ValueLookupParameter<IOperator>("EmigrantsSelector", "Selects the individuals that will be migrated."));
      Parameters.Add(new ValueLookupParameter<IOperator>("ImmigrationReplacer", "Replaces some of the original population with the immigrants."));
      Parameters.Add(new ValueLookupParameter<IntValue>("PopulationSize", "The size of the population of solutions."));
      Parameters.Add(new ValueLookupParameter<IntValue>("CurrentPopulationSize", "The current size of the population of solutions."));

      Parameters.Add(new ValueLookupParameter<IOperator>("Selector", "The operator used to select solutions for reproduction."));
      Parameters.Add(new ValueLookupParameter<IOperator>("Crossover", "The operator used to cross solutions."));
      Parameters.Add(new ValueLookupParameter<PercentValue>("MutationProbability", "The probability that the mutation operator is applied on a solution."));
      Parameters.Add(new ValueLookupParameter<IOperator>("Mutator", "The operator used to mutate solutions."));
      Parameters.Add(new ValueLookupParameter<IOperator>("Evaluator", "The operator used to evaluate solutions."));
      Parameters.Add(new ValueLookupParameter<IntValue>("Elites", "The numer of elite solutions which are kept in each generation."));
      Parameters.Add(new ValueLookupParameter<BoolValue>("ReevaluateElites", "Flag to determine if elite individuals should be reevaluated (i.e., if stochastic fitness functions are used.)"));
      Parameters.Add(new ValueLookupParameter<ResultCollection>("Results", "The results collection to store the results."));
      Parameters.Add(new ValueLookupParameter<ITerminator>("Terminator", "The termination criteria that defines if the algorithm should continue or stop"));

      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Age", "The age of individuals."));
      Parameters.Add(new ValueLookupParameter<IntValue>("AgeGap", "The frequency of reseeding the lowest layer and scaling factor for the age-limits for the layers."));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("AgeInheritance", "A weight that determines the age of a child after crossover based on the older (1.0) and younger (0.0) parent."));
      Parameters.Add(new ValueLookupParameter<IntArray>("AgeLimits", "The maximum age an individual is allowed to reach in a certain layer."));
      Parameters.Add(new ValueLookupParameter<BoolValue>("PlusSelection", "Include the parents in the selection of the invividuals for the next generation."));
      Parameters.Add(new ValueLookupParameter<IntValue>("MatingPoolRange", "The range of sub - populations used for creating a mating pool. (1 = current + previous sub-population)"));
      Parameters.Add(new ValueLookupParameter<BoolValue>("ReduceToPopulationSize", "Reduce the CurrentPopulationSize after elder migration to PopulationSize"));

      Parameters.Add(new ValueLookupParameter<IOperator>("Analyzer", "The operator used to the analyze the islands."));
      Parameters.Add(new ValueLookupParameter<IOperator>("IslandAnalyzer", "The operator used to analyze each island."));
      Parameters.Add(new ValueLookupParameter<IOperator>("LayerAnalyzer", "The operator used to analyze each Layer."));
      Parameters.Add(new ValueLookupParameter<IOperator>("GroupAnalyzer", "The operator used to analyze each Layergroup."));

      Parameters.Add(new LookupParameter<IntValue>("LayerEvaluatedSolutions", "The number of times a solution has been evaluated."));
      //TODO: make long ?
      Parameters.Add(new LookupParameter<IntValue>("IslandEvaluatedSolutions", "The number of times a solution has been evaluated on one island."));
      Parameters.Add(new LookupParameter<IntValue>("EvaluatedSolutions", "The number of times a solution has been evaluated on one island."));

      Parameters.Add(new ValueLookupParameter<BoolValue>("Migrate", "Migrate the island?"));

      #endregion
      var globalVariableCreator = new VariableCreator() { Name = "Create global values" };
      var islandVarCreatorUssp = new UniformSubScopesProcessor();
      var islandVariableCreator = new VariableCreator() {Name = "Initialize Results"};
      var groupVariableCreator = new VariableCreator() {Name = "Initialize Group Results"};

      var layerUssp = new UniformSubScopesProcessor();
      var layerVariableCreator = new VariableCreator() {};

      var initLayerAnalyzerPlaceholder = new Placeholder() {Name = "Layer Analyzer (Placeholder)"};
      var initIslandAnalyzerPlaceholder = new Placeholder() {Name = "Island Analyzer (Placeholder)"};
      var initGlobalAnalyzerPlacerholder =new Placeholder() {Name = "Global Analyzer (Placeholder)"};
      var resultsCollector = new ResultsCollector();

      var processIslandsUssp = new UniformSubScopesProcessor() {Name = "Process Islands", Parallel = new BoolValue(true)};
      var initIslandEvaluationsAss = new Assigner() {Name = "Init Island Evaluations"};

      var incrementGenerationDr = new IntCounter() { Name = "Increment Generations" };
      var incrementEvaluatedSolutionDr = new DataReducer() { Name = "Increment Evaluated Solutions" };

      var elderMigrator = CreateEldersEmigrator();
      var newLayerOpener = CreateLayerOpener();
      var reseeder = CreateReseeder();

      var analyzeLayerUssp = new UniformSubScopesProcessor();

      var layerAnalyzerPlaceholder = new Placeholder() { Name = "Layer Analyzer (Placeholder)" };
      var layerResultsCollector = new ResultsCollector() {Name = "Collect Additional Infos"};
      var islandAnalyzerPlaceholder = new Placeholder() { Name = "Island Analyzer (Placeholder)" };
      var islandResultsCollector = new ResultsCollector() { Name = "Collect layer results into IslandResults" };

      var globalAnalyzerPlacerholder = new Placeholder() { Name = "Global Analyzer (Placeholder)" };

      var groupingOperator = new GroupedLayerOperator() {Name = "Grouped Analyzer"};
      var groupAnalyzer = new Placeholder() {Name = "Group Analyzer (Placeholder)"};
      var groupResultsExtractor = new ResultsExtractor() {Name = "Collect group results into variable"};
      var groupResultsCollector = new ResultsCollector() {Name = "Collect group results into into global results"};
      var groupLayerNumberResultsCollector = new ResultsCollector() {Name = "Add Layernumber to Results"};

      var matingPoolCreator = new MatingPoolCreator();
      var matingPoolUssp = new UniformSubScopesProcessor() {Name = "Process Mating Pools"};
      var alpsMainOperator = new AlpsGeneticAlgorithmMainOperator();

      var setIslandEvaluatedSolutions = new DataReducer() {Name = "Set IslandEvaluatedSolutions" };

      var migrationThresholdCounter = new IntCounter() {Name = "Increment Counter for Migrations"};
      var migrationThresholdAss = new Assigner() {Name = "Reset Counter for Migrations"};
      var migrationCounter = new IntCounter() {Name = "Increment  Migrations"};
      var selectforMigrationUssp = new UniformSubScopesProcessor();
      var emigrantSelector = new Placeholder() {Name = "Emigrant Selector (Placeholder)"};
      var migrator = new LayerMigrator();
      var replaceforMigrationUssp = new UniformSubScopesProcessor();
      var immigrantSelector = new Placeholder() { Name = "Immigrant Replacer (Placeholder)" };

      var migrateComparator = new Comparator();
      var migrateCondBranch = new ConditionalBranch() {Name = "Migrate?"};

      var terminator = new TerminationOperator();
      
      globalVariableCreator.CollectedValues.Add(new ValueParameter<IntValue>("Generations",new IntValue(0)));
      globalVariableCreator.CollectedValues.Add(new ValueParameter<IntValue>("Migrations",new IntValue(0)));

      islandVariableCreator.CollectedValues.Add(new ValueParameter<IntValue>("OpenLayers", new IntValue(1)));
      islandVariableCreator.CollectedValues.Add(new ValueParameter<ResultCollection>("IslandResults"));

      layerVariableCreator.CollectedValues.Add(new ValueParameter<IntValue>("Layer", new IntValue(0)));
      layerVariableCreator.CollectedValues.Add(new ValueParameter<ResultCollection>("LayerResults"));

      groupVariableCreator.CollectedValues.Add((new ValueParameter<ResultCollection>("GroupResults")));
      //TODO: groupAnalyze
      initLayerAnalyzerPlaceholder.OperatorParameter.ActualName = LayerAnalyzerParameter.Name;
      initIslandAnalyzerPlaceholder.OperatorParameter.ActualName = IslandAnalyzerParameter.Name;
      initGlobalAnalyzerPlacerholder.OperatorParameter.ActualName = AnalyzerParameter.Name;

      resultsCollector.CollectedValues.Add(new LookupParameter<IntValue>("Generations"));
      resultsCollector.CollectedValues.Add(new ScopeTreeLookupParameter<ResultCollection>("IslandResults","","IslandResults"));
      resultsCollector.CollectedValues.Add(new ScopeTreeLookupParameter<ResultCollection>("GroupResults","","GroupResults"));
      resultsCollector.CollectedValues.Add(new ValueLookupParameter<IntValue>("Migrations"));
      resultsCollector.CopyValue = new BoolValue(false);

      initIslandEvaluationsAss.LeftSideParameter.ActualName = IslandEvaluatedSolutions.Name;
      initIslandEvaluationsAss.RightSideParameter.Value = new IntValue(0);

      incrementGenerationDr.ValueParameter.ActualName = "Generations";

      incrementEvaluatedSolutionDr.ParameterToReduce.ActualName = IslandEvaluatedSolutions.Name;
      incrementEvaluatedSolutionDr.TargetParameter.ActualName = EvaluatedSolutions.Name;
      incrementEvaluatedSolutionDr.ReductionOperation.Value = new ReductionOperation(ReductionOperations.Sum);
      incrementEvaluatedSolutionDr.TargetOperation.Value = new ReductionOperation(ReductionOperations.Sum);

      setIslandEvaluatedSolutions.ParameterToReduce.ActualName = LayerEvaluatedSolutionsParameter.Name;
      setIslandEvaluatedSolutions.ReductionOperation.Value = new ReductionOperation(ReductionOperations.Sum);
      setIslandEvaluatedSolutions.TargetOperation.Value = new ReductionOperation(ReductionOperations.Assign);
      setIslandEvaluatedSolutions.TargetParameter.ActualName = IslandEvaluatedSolutions.Name;


      migrationThresholdCounter.ValueParameter.ActualName = "MigrationThreshold";
      
      migrateComparator.Comparison.Value = ComparisonType.GreaterOrEqual;
      migrateComparator.LeftSideParameter.ActualName = "MigrationThreshold";
      migrateComparator.RightSideParameter.ActualName = MigrationIntervalParameter.Name;
      migrateComparator.ResultParameter.ActualName = "Migrate";

      migrateCondBranch.ConditionParameter.ActualName = "Migrate";

      migrationCounter.ValueParameter.ActualName = "Migrations";
      migrationThresholdAss.LeftSideParameter.ActualName = "MigrationThreshold";
      migrationThresholdAss.RightSideParameter.Value = new IntValue(0);

      selectforMigrationUssp.Depth = new IntValue(2);
      emigrantSelector.OperatorParameter.ActualName = EmigrantsSelectorParameter.Name;

      replaceforMigrationUssp.Depth = new IntValue(2);
      immigrantSelector.OperatorParameter.ActualName = ImmigrationReplacerParameter.Name;

      groupAnalyzer.OperatorParameter.ActualName = GroupAnalyzerParameter.Name;
      groupLayerNumberResultsCollector.CollectedValues.Add(new LookupParameter<IntValue>("LayerNumber"));
      groupLayerNumberResultsCollector.ResultsParameter.ActualName = "GroupResults";

      layerAnalyzerPlaceholder.OperatorParameter.ActualName = LayerAnalyzerParameter.Name;
      layerResultsCollector.ResultsParameter.ActualName = "LayerResults";
      layerResultsCollector.CollectedValues.Add(new LookupParameter<IntValue>("Layer"));
      islandAnalyzerPlaceholder.OperatorParameter.ActualName = IslandAnalyzerParameter.Name;
      islandResultsCollector.ResultsParameter.ActualName = "IslandResults";
      islandResultsCollector.CollectedValues.Add(new ScopeTreeLookupParameter<ResultCollection>("LayerResults"));
      globalAnalyzerPlacerholder.OperatorParameter.ActualName = AnalyzerParameter.Name;

      //TODO: implement check for grouped layer analyzer
      groupResultsExtractor.SourceNameParameter.Value = new StringValue("GroupScope");
      groupResultsExtractor.TargetNameParameter.Value = new StringValue("GroupResults");
      groupResultsCollector.CollectedValues.Add(new LookupParameter<ItemArray<ResultCollection>>("GroupResults"));
      matingPoolCreator.MatingPoolRangeParameter.ActualName = MatingPoolRangeParameter.Name; 

      alpsMainOperator.RandomParameter.ActualName = LayerRandomParameter.Name;
      alpsMainOperator.EvaluatorParameter.ActualName = EvaluatorParameter.Name;
      alpsMainOperator.EvaluatedSolutionsParameter.ActualName = LayerEvaluatedSolutionsParameter.Name;
      alpsMainOperator.QualityParameter.ActualName = QualityParameter.Name;
      alpsMainOperator.MaximizationParameter.ActualName = MaximizationParameter.Name;
      alpsMainOperator.PopulationSizeParameter.ActualName = PopulationSizeParameter.Name;
      alpsMainOperator.SelectorParameter.ActualName = SelectorParameter.Name;
      alpsMainOperator.CrossoverParameter.ActualName = CrossoverParameter.Name;
      alpsMainOperator.MutatorParameter.ActualName = MutatorParameter.ActualName;
      alpsMainOperator.MutationProbabilityParameter.ActualName = MutationProbabilityParameter.Name;
      alpsMainOperator.ElitesParameter.ActualName = ElitesParameter.Name;
      alpsMainOperator.ReevaluateElitesParameter.ActualName = ReevaluateElitesParameter.Name;
      alpsMainOperator.PlusSelectionParameter.ActualName = PlusSelectionParameter.Name;
      alpsMainOperator.AgeParameter.ActualName = AgeParameter.Name;
      alpsMainOperator.AgeInheritanceParameter.ActualName = AgeInheritanceParameter.Name;
      alpsMainOperator.AgeIncrementParameter.Value = new DoubleValue(1.0);

      //flow
      OperatorGraph.InitialOperator = globalVariableCreator;
      globalVariableCreator.Successor = islandVarCreatorUssp;
      islandVarCreatorUssp.Operator = islandVariableCreator;
      islandVariableCreator.Successor = layerUssp;
      layerUssp.Operator = layerVariableCreator;
      layerVariableCreator.Successor = initLayerAnalyzerPlaceholder;
      initLayerAnalyzerPlaceholder.Successor = null;
      layerUssp.Successor = initIslandAnalyzerPlaceholder;
      initLayerAnalyzerPlaceholder.Successor = null;
      islandVarCreatorUssp.Successor = initGlobalAnalyzerPlacerholder;
      initGlobalAnalyzerPlacerholder.Successor = resultsCollector;
      resultsCollector.Successor = processIslandsUssp;
      processIslandsUssp.Operator = initIslandEvaluationsAss;
      initIslandEvaluationsAss.Successor = matingPoolCreator;
      matingPoolCreator.Successor = matingPoolUssp;
      matingPoolUssp.Operator = alpsMainOperator;
      alpsMainOperator.Successor = null;
      matingPoolUssp.Successor = elderMigrator;
      elderMigrator.Successor = newLayerOpener;
      newLayerOpener.Successor = reseeder;
      reseeder.Successor = analyzeLayerUssp;
      analyzeLayerUssp.Operator = layerAnalyzerPlaceholder;
      layerAnalyzerPlaceholder.Successor = layerResultsCollector;
      layerResultsCollector.Successor = null;

      analyzeLayerUssp.Successor = islandAnalyzerPlaceholder;
      islandAnalyzerPlaceholder.Successor = islandResultsCollector;
      islandResultsCollector.Successor = setIslandEvaluatedSolutions;
      setIslandEvaluatedSolutions.Successor = null;

      processIslandsUssp.Successor = incrementGenerationDr;
      incrementGenerationDr.Successor = incrementEvaluatedSolutionDr;
      incrementEvaluatedSolutionDr.Successor = migrationThresholdCounter;

      migrationThresholdCounter.Successor = migrateComparator;

      migrateComparator.Successor = migrateCondBranch;
      migrateCondBranch.FalseBranch = null;
      migrateCondBranch.TrueBranch = migrationThresholdAss;
      migrationThresholdAss.Successor = migrationCounter;

      migrationCounter.Successor = selectforMigrationUssp;
      selectforMigrationUssp.Operator = emigrantSelector;
      emigrantSelector.Successor = null;
      selectforMigrationUssp.Successor = migrator;

      migrator.Successor = replaceforMigrationUssp;
      replaceforMigrationUssp.Operator = immigrantSelector;
      immigrantSelector.Successor = null;
      replaceforMigrationUssp.Successor = null;

      migrateCondBranch.Successor = globalAnalyzerPlacerholder;


      globalAnalyzerPlacerholder.Successor = groupingOperator;
      groupingOperator.Operator = groupAnalyzer;
      groupAnalyzer.Successor = groupLayerNumberResultsCollector;
      groupLayerNumberResultsCollector.Successor = null;
      groupingOperator.Successor = groupResultsExtractor;
      groupResultsExtractor.Successor = groupResultsCollector;
      groupResultsCollector.Successor = terminator;
      terminator.ContinueBranch = processIslandsUssp;

    }

    private CombinedOperator CreateEldersEmigrator() {
      var eldersEmigrator = new CombinedOperator() { Name = "Emigrate Elders" };
      var selectorProsessor = new UniformSubScopesProcessor();
      var eldersSelector = new EldersSelector();
      var shiftToRightMigrator = new UnidirectionalRingMigrator() { Name = "Shift elders to next layer" };
      var mergingProsessor = new UniformSubScopesProcessor();
      var mergingReducer = new MergingReducer();
      var subScopesCounter = new SubScopesCounter();
      var reduceToPopulationSizeBranch = new ConditionalBranch() { Name = "ReduceToPopulationSize?" };
      var countCalculator = new ExpressionCalculator() { Name = "CurrentPopulationSize = Min(CurrentPopulationSize, PopulationSize)" };
      var bestSelector = new BestSelector();
      var rightReducer = new RightReducer();

      eldersEmigrator.OperatorGraph.InitialOperator = selectorProsessor;

      selectorProsessor.Operator = eldersSelector;
      selectorProsessor.Successor = shiftToRightMigrator;

      eldersSelector.AgeParameter.ActualName = AgeParameter.Name;
      eldersSelector.AgeLimitsParameter.ActualName = AgeLimitsParameter.Name;
      eldersSelector.NumberOfLayersParameter.ActualName = NumberOfLayersParameter.Name;
      eldersSelector.LayerParameter.ActualName = "Layer";
      eldersSelector.Successor = null;

      shiftToRightMigrator.ClockwiseMigrationParameter.Value = new BoolValue(true);
      shiftToRightMigrator.Successor = mergingProsessor;

      mergingProsessor.Operator = mergingReducer;

      mergingReducer.Successor = subScopesCounter;

      subScopesCounter.ValueParameter.ActualName = CurrentPopulationSizeParameter.Name;
      subScopesCounter.AccumulateParameter.Value = new BoolValue(false);
      subScopesCounter.Successor = reduceToPopulationSizeBranch;

      reduceToPopulationSizeBranch.ConditionParameter.ActualName = ReduceToPopulationSizeParameter.Name;
      reduceToPopulationSizeBranch.TrueBranch = countCalculator;

      countCalculator.CollectedValues.Add(new LookupParameter<IntValue>(PopulationSizeParameter.Name));
      countCalculator.CollectedValues.Add(new LookupParameter<IntValue>(CurrentPopulationSizeParameter.Name));
      countCalculator.ExpressionParameter.Value = new StringValue("CurrentPopulationSize PopulationSize CurrentPopulationSize PopulationSize < if toint");
      countCalculator.ExpressionResultParameter.ActualName = CurrentPopulationSizeParameter.Name;
      countCalculator.Successor = bestSelector;

      bestSelector.NumberOfSelectedSubScopesParameter.ActualName = CurrentPopulationSizeParameter.Name;
      bestSelector.CopySelected = new BoolValue(false);
      bestSelector.Successor = rightReducer;

      return eldersEmigrator;
    }

    private CombinedOperator CreateLayerOpener() {
      var layerOpener = new CombinedOperator() { Name = "Open new Layer if needed" };
      var maxLayerReached = new Comparator() { Name = "MaxLayersReached = OpenLayers >= NumberOfLayers" };
      var maxLayerReachedBranch = new ConditionalBranch() { Name = "MaxLayersReached?" };
      var openNewLayerCalculator = new ExpressionCalculator() { Name = "OpenNewLayer = Generations >= AgeLimits[OpenLayers - 1]" };
      var openNewLayerBranch = new ConditionalBranch() { Name = "OpenNewLayer?" };
      var layerCreator = new LastLayerCloner() { Name = "Create Layer" };
      var updateLayerNumber = new Assigner() { Name = "Layer = OpenLayers" };
      var historyWiper = new ResultsHistoryWiper() { Name = "Clear History in Results" };
      var createChildrenViaCrossover = new AlpsGeneticAlgorithmMainOperator();
      var incrEvaluatedSolutionsForNewLayer = new SubScopesCounter() { Name = "Update EvaluatedSolutions" };
      var incrOpenLayers = new IntCounter() { Name = "Incr. OpenLayers" };
      //var newLayerResultsCollector = new ResultsCollector() { Name = "Collect new Layer Results" };

      layerOpener.OperatorGraph.InitialOperator = maxLayerReached;

      maxLayerReached.LeftSideParameter.ActualName = "OpenLayers";
      maxLayerReached.RightSideParameter.ActualName = NumberOfLayersParameter.Name;
      maxLayerReached.ResultParameter.ActualName = "MaxLayerReached";
      maxLayerReached.Comparison = new Comparison(ComparisonType.GreaterOrEqual);
      maxLayerReached.Successor = maxLayerReachedBranch;

      maxLayerReachedBranch.ConditionParameter.ActualName = "MaxLayerReached";
      maxLayerReachedBranch.FalseBranch = openNewLayerCalculator;

      openNewLayerCalculator.CollectedValues.Add(new LookupParameter<IntArray>(AgeLimitsParameter.Name));
      openNewLayerCalculator.CollectedValues.Add(new LookupParameter<IntValue>("Generations"));
      openNewLayerCalculator.CollectedValues.Add(new LookupParameter<IntValue>(NumberOfLayersParameter.Name));
      openNewLayerCalculator.CollectedValues.Add(new LookupParameter<IntValue>("OpenLayers"));
      openNewLayerCalculator.ExpressionResultParameter.ActualName = "OpenNewLayer";
      openNewLayerCalculator.ExpressionParameter.Value = new StringValue("Generations 1 + AgeLimits OpenLayers 1 - [] >");
      openNewLayerCalculator.Successor = openNewLayerBranch;

      openNewLayerBranch.ConditionParameter.ActualName = "OpenNewLayer";
      openNewLayerBranch.TrueBranch = layerCreator;

      layerCreator.NewLayerOperator = updateLayerNumber;
      layerCreator.Successor = incrOpenLayers;

      updateLayerNumber.LeftSideParameter.ActualName = "Layer";
      updateLayerNumber.RightSideParameter.ActualName = "OpenLayers";
      updateLayerNumber.Successor = historyWiper;

      historyWiper.ResultsParameter.ActualName = "LayerResults";
      historyWiper.Successor = createChildrenViaCrossover;

      // Maybe use only crossover and no elitism instead of "default operator"
      createChildrenViaCrossover.RandomParameter.ActualName = LayerRandomParameter.Name;
      createChildrenViaCrossover.EvaluatorParameter.ActualName = EvaluatorParameter.Name;
      createChildrenViaCrossover.EvaluatedSolutionsParameter.ActualName = "LayerEvaluatedSolutions";
      createChildrenViaCrossover.QualityParameter.ActualName = QualityParameter.Name;
      createChildrenViaCrossover.MaximizationParameter.ActualName = MaximizationParameter.Name;
      createChildrenViaCrossover.PopulationSizeParameter.ActualName = PopulationSizeParameter.Name;
      createChildrenViaCrossover.SelectorParameter.ActualName = SelectorParameter.Name;
      createChildrenViaCrossover.CrossoverParameter.ActualName = CrossoverParameter.Name;
      createChildrenViaCrossover.MutatorParameter.ActualName = MutatorParameter.Name;
      createChildrenViaCrossover.MutationProbabilityParameter.ActualName = MutationProbabilityParameter.Name;
      createChildrenViaCrossover.ElitesParameter.ActualName = ElitesParameter.Name;
      createChildrenViaCrossover.ReevaluateElitesParameter.ActualName = ReevaluateElitesParameter.Name;
      createChildrenViaCrossover.PlusSelectionParameter.ActualName = PlusSelectionParameter.Name;
      createChildrenViaCrossover.AgeParameter.ActualName = AgeParameter.Name;
      createChildrenViaCrossover.AgeInheritanceParameter.ActualName = AgeInheritanceParameter.Name;
      createChildrenViaCrossover.AgeIncrementParameter.Value = new DoubleValue(0.0);
      createChildrenViaCrossover.Successor = incrEvaluatedSolutionsForNewLayer;

      incrEvaluatedSolutionsForNewLayer.ValueParameter.ActualName = LayerEvaluatedSolutionsParameter.Name;
      incrEvaluatedSolutionsForNewLayer.AccumulateParameter.Value = new BoolValue(true);

      incrOpenLayers.ValueParameter.ActualName = "OpenLayers";
      incrOpenLayers.Increment = new IntValue(1);

      return layerOpener;
    }

    private CombinedOperator CreateReseeder() {
      var reseeder = new CombinedOperator() { Name = "Reseed Layer Zero if needed" };
      var reseedingController = new ReseedingController() { Name = "Reseeding needed (Generation % AgeGap == 0)?" };
      var removeIndividuals = new SubScopesRemover();
      var createIndividuals = new SolutionsCreator();
      var initializeAgeProsessor = new UniformSubScopesProcessor();
      var initializeAge = new VariableCreator() { Name = "Initialize Age" };
      var incrEvaluatedSolutionsAfterReseeding = new SubScopesCounter() { Name = "Update EvaluatedSolutions" };

      reseeder.OperatorGraph.InitialOperator = reseedingController;

      reseedingController.GenerationsParameter.ActualName = "Generations";
      reseedingController.AgeGapParameter.ActualName = AgeGapParameter.Name;
      reseedingController.FirstLayerOperator = removeIndividuals;
      reseedingController.Successor = null;

      removeIndividuals.Successor = createIndividuals;

      createIndividuals.NumberOfSolutionsParameter.ActualName = PopulationSizeParameter.Name;
      createIndividuals.Successor = initializeAgeProsessor;

      initializeAgeProsessor.Operator = initializeAge;
      initializeAgeProsessor.Successor = incrEvaluatedSolutionsAfterReseeding;

      initializeAge.CollectedValues.Add(new ValueParameter<DoubleValue>(AgeParameter.Name, new DoubleValue(0)));

      incrEvaluatedSolutionsAfterReseeding.ValueParameter.ActualName = LayerEvaluatedSolutionsParameter.Name;
      incrEvaluatedSolutionsAfterReseeding.AccumulateParameter.Value = new BoolValue(true);

      return reseeder;
    }
  }
}
